/* ***** BEGIN LICENSE BLOCK *****
 *   Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is Save File to.
 *
 * The Initial Developer of the Original Code is
 * Atte Kemppilä.
 * Portions created by the Initial Developer are Copyright (C) 2007
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 * 
 * ***** END LICENSE BLOCK ***** */


var savefileto =
{
    doOnLoad: function(e)
    {
        this.initialized = true;
        this.prefs = savefileto_util.getPrefBranch("extensions.savefileto.");
        // showPredicates are the same as in browser.jar:content/browser/browser.js:nsContextMenu.initSaveItems()
        // XXX: ff3.1: onCanvas not supported
        this.image = this.createSaveFileTo("image", function() { return gContextMenu.onLoadedImage; }, savefileto_saveImage);
        this.link = this.createSaveFileTo("link", function() { return gContextMenu.onSaveableLink; }, savefileto_saveLink);
        this.page = this.createSaveFileTo("page", function() { return (! (gContextMenu.inDirList || gContextMenu.isContentSelected || gContextMenu.onTextInput || gContextMenu.onLink || gContextMenu.onImage || gContextMenu.onCanvas || gContextMenu.onVideo || gContextMenu.onAudio)); }, savefileto_savePage);
        var newVersion = savefileto_util.getExtensionVersion("savefileto@mozdev.org");
        var oldVersion = this.prefs.getCharPref("lastVersion");
        if (savefileto_util.compareVersion(newVersion, oldVersion) != 0)
            this.initVersion(newVersion, oldVersion);
    },
    
    // TODO: move to component
    initVersion: function(newVersion, oldVersion)
    {
        var l10n = document.getElementById("savefileto-l10n");
        // first init ever
        if (oldVersion == "0")
        {
            // defaults cannot be made platform specific - let's do it by hand
            // TODO: better defaults for Mac
            if (savefileto_util.isWin)
                var folders = ["Desk", "Pers"];
            else
                var folders = ["Desk", "Home"];
            for (var i = 0; i < folders.length; i++)
            {
                var name = l10n.getString("specialFolder." + folders[i]);
                var path = savefileto_util.getSpecialFile(folders[i]).path;
                folders[i] = name + "*" + path;
            }
            folders = folders.join("|");
            // it would be nice to set these into the default branch,
            // but that doesn't seem to persisted :(
            savefileto_util.setUnicharPref(this.prefs, "page.rootFolders", folders);
            savefileto_util.setUnicharPref(this.prefs, "link.rootFolders", folders);
            savefileto_util.setUnicharPref(this.prefs, "image.rootFolders", folders);
        }
        this.prefs.setCharPref("lastVersion", newVersion);
    },
    
    createSaveFileTo: function(name, showPredicate, doOnAction)
    {
        var Name = name.charAt(0).toUpperCase() + name.substr(1);
        var l10n = document.getElementById("savefileto-l10n");
        var prefsName = "extensions.savefileto.";
        var folderPrefPrefix = name + ".";
        var menuItemId = "savefileto-contextSave" + Name + "To";
        var mainMenuId = "savefileto-" + name + "MainMenu";
        var actionLabel = l10n.getString("saveHere");
        return new savefileto_SubFolderMenu(prefsName, folderPrefPrefix, menuItemId, mainMenuId, showPredicate, actionLabel, doOnAction);
    }
};


function savefileto_SubFolderMenu(prefsName, folderPrefPrefix, menuItemId, mainMenuId, showPredicate, actionLabel, doOnAction)
{
    this.prefs = savefileto_util.getPrefBranch(prefsName);
    this.folderPrefPrefix = folderPrefPrefix;
    this.menuItemId = menuItemId;
    this.mainMenuId = mainMenuId;
    this.showPredicate = showPredicate;
    this.actionLabel = actionLabel;
    this.doOnAction = doOnAction;
    var callback = savefileto_util.getEventHandler(this, "doOnContextMenuShowing")
    document.getElementById("contentAreaContextMenu").addEventListener("popupshowing", callback, false);
}

savefileto_SubFolderMenu.prototype =
{
    get enabled()        { return this.prefs.getBoolPref(this.folderPrefPrefix + "enabled"); },
    get lookahead()      { return this.prefs.getBoolPref("lookahead"); },
    get maxMenuLength()  { return this.prefs.getIntPref("maxMenuLength"); },
    get recentFoldersN() { return this.prefs.getIntPref("recentFoldersN"); },
    get recentFoldersOnTop() { return this.prefs.getBoolPref("recentFoldersOnTop"); },
    get prompt()         { return this.prefs.getBoolPref("prompt"); },
    
    doOnContextMenuShowing: function(e)
    {
        var show = (this.enabled && this.showPredicate());
        gContextMenu.showItem(this.menuItemId, show);
    },
    
    doOnMainMenuShowing: function(e)
    {
        e.stopPropagation();
        var menu = e.originalTarget;
        this.clearMenu(menu);
        this.populateMainMenu(menu);
    },
    
    doOnMainMenuHidden: function(e)
    {
        e.stopPropagation();
        var menu = e.originalTarget;
        this.clearMenu(menu);
    },
    
    doOnSubMenuShowing: function(e, folder)
    {
        e.stopPropagation();
        var menu = e.originalTarget;
        if (menu.childNodes.length == 0)
            this.populateSubMenu(menu, folder);
    },
    
    doOnSubMenuHidden: function(e)
    {
        e.stopPropagation();
    },
    
    clearMenu: function(menu)
    {
        var childNodes = menu.childNodes;
        for (var i = childNodes.length - 1; i >= 0; i--)
            menu.removeChild(childNodes[i]);
    },
    
    populateMainMenu: function(menu)
    {
        var rootFolders = savefileto_util.getUnicharPref(this.prefs, this.folderPrefPrefix + "rootFolders");
        rootFolders = savefileto_util.parseFolders(rootFolders, true);
        var recentFolders = this.getRecentFolders();
        
        this.appendRecentFolders(menu, recentFolders, true);
        for (var i = 0; i < rootFolders.length; i++)
            if (rootFolders[i].file.exists() && rootFolders[i].file.isDirectory())
                this.appendSubFolderMenu(menu, rootFolders[i].name, rootFolders[i]);
        this.appendRecentFolders(menu, recentFolders, false);
    },
    
    getRecentFolders: function()
    {
        var recentFolders = savefileto_util.getUnicharPref(this.prefs, this.folderPrefPrefix + "recentFolders");
        recentFolders = savefileto_util.parseFolders(recentFolders, true);
        var n = Math.min(this.recentFoldersN, recentFolders.length);
        var folders = [];
        for (var i = 0; i < n; i++)
            if (recentFolders[i].file.exists() && recentFolders[i].file.isDirectory())
                folders.push(recentFolders[i]);
        return folders;
    },
    
    appendRecentFolders: function(menu, folders, top)
    {
        if ((folders.length > 0) && (this.recentFoldersOnTop == top))
        {
            if (! top)
                savefileto_util.appendMenuSeparator(menu);
            for (var i = 0; i < folders.length; i++)
                this.appendFolderMenuItem(menu, folders[i].name, folders[i]);
            if (top)
                savefileto_util.appendMenuSeparator(menu);
        }
    },
    
    populateSubMenu: function(menu, folder)
    {
        if (savefileto_util.isArray(folder))
        {
            var subFolderList = [];
            for (var i = 0; i < folder.length; i++)
                subFolderList.push({prefix: folder[i].leafName.toLowerCase(), folders: [folder[i]]});
        }
        else
        {
            var subFolderList = this.getSubFolderList(folder);
            this.appendFolderMenuItem(menu, this.actionLabel, folder);
            if (subFolderList.length == 0)
                return;
            savefileto_util.appendMenuSeparator(menu);
        }
        if (this.maxMenuLength > 1)
            subFolderList = this.compactFolderList(subFolderList, this.maxMenuLength);
        for (var i = 0; i < subFolderList.length; i++)
        {
            var entry = subFolderList[i];
            var single = (entry.folders.length == 1);
            var subFolder = (single ? entry.folders[0] : entry.folders);
            var name = (single ? subFolder.leafName : entry.prefix + "...");
            this.appendSubFolderMenu(menu, name, subFolder);
        }
    },
    
    getSubFolderList: function(folder, findOnly)
    {
        const reWinShortcut = /.+\.lnk$/;
        var subFolderList = [];
        var entries = folder.file.directoryEntries;
        while (entries.hasMoreElements())
        {
            var file = entries.getNext();
            file.QueryInterface(Components.interfaces.nsIFile);
            var name = file.leafName;
            // ignore windows shortcuts
            if (savefileto_util.isWin && reWinShortcut.test(name))
                continue;
            if (! file.isDirectory())
                continue;
            if ((name.charAt(0) == ".") || file.isHidden())
                continue;
            if (findOnly)
                return true;
            var subFolder = {};
            subFolder.file = file;
            subFolder.path = file.path;
            subFolder.name = folder.name + savefileto_util.pathSep + name;
            subFolder.leafName = name;
            subFolderList.push({prefix: name.toLowerCase(), folders: [subFolder]});
        }
        if (findOnly)
            return false;
        subFolderList.sort(function(a, b) { return savefileto_util.compare(a.prefix, b.prefix); });
        return subFolderList;
    },
    
    hasSubFolders: function(folder)
    {
        return this.getSubFolderList(folder, true);
    },
    
    compactFolderList: function(folderList, max)
    {
        if (folderList.length <= max)
            return folderList;
        var reNonAlphaNumEnd = /(.)[^A-Za-z0-9]+$/;
        var trim = function(s) { return s.replace(reNonAlphaNumEnd, "$1"); };
        var list = [folderList[0]];
        var comparator = function(a, b) { var d = a.length - b.length; return ((d != 0) ? d : savefileto_util.compare(b, a)); };
        var prefixHeap = new savefileto_Heap(comparator);
        for (var i = 1; i < folderList.length; i++)
        {
            var entry = folderList[i];
            var lastEntry = list[list.length - 1];
            var prefix = trim(savefileto_util.commonPrefix(lastEntry.prefix, entry.prefix));
            if ((lastEntry.folders.length > 1) && (prefix == lastEntry.prefix))
                lastEntry.folders.push(entry.folders[0]);
            else
            {
                list.push(entry);
                prefixHeap.push(prefix);
                if (list.length > max)
                {
                    prefix = prefixHeap.pop();
                    if (prefix.length > 0)
                    {
                        var lo = 0;
                        while ((lo < list.length) && (list[lo].prefix.substr(0, prefix.length) != prefix))
                            lo++;
                        var hi = lo + 1;
                        while ((hi < list.length) && (list[hi].prefix.substr(0, prefix.length) == prefix))
                            hi++;
                        var n = hi - lo;
                        var subList = list.splice(lo, n);
                        var subFolders = [];
                        for (var k = 0; k < subList.length; k++)
                            subFolders = subFolders.concat(subList[k].folders);
                        list.splice(lo, 0, {prefix: prefix, folders: subFolders});
                        while ((prefixHeap.length > 0) && (prefixHeap.peek() == prefix))
                            prefixHeap.pop();
                        if (lo > 0)
                            prefixHeap.push(trim(savefileto_util.commonPrefix(list[lo - 1].prefix, list[lo].prefix)));
                        if (lo + 1 < list.length)
                            prefixHeap.push(trim(savefileto_util.commonPrefix(list[lo + 1].prefix, list[lo].prefix)));
                    }
                }
            }
        }
        if (list.length == 1)
            return folderList;
        return list;
    },
    
    appendSubFolderMenu: function(menu, label, folder)
    {
        if ((! savefileto_util.isArray(folder)) && this.lookahead && (! this.hasSubFolders(folder)))
            this.appendFolderMenuItem(menu, label, folder);
        else
        {
            var subMenu = document.createElement("menu");
            subMenu.setAttribute("label", label);
            var menuPopup = document.createElement("menupopup");
            var callback = savefileto_util.getEventHandler(this, "doOnSubMenuShowing", true, folder);
            menuPopup.addEventListener("popupshowing", callback, false);
            callback = savefileto_util.getEventHandler(this, "doOnSubMenuHidden");
            menuPopup.addEventListener("popuphidden", callback, false);
            menuPopup.className = "savefileto-subMenu";
            subMenu.appendChild(menuPopup);
            menu.appendChild(subMenu);
        }
    },
    
    appendFolderMenuItem: function(menu, label, folder)
    {
        var callback = savefileto_util.getEventHandler(this, "doOnAction", true, folder);
        savefileto_util.appendMenuItem(menu, label, callback);
    },
    
    addToRecentFolders: function(folder)
    {
        var recentFolders = [];
        if (this.recentFoldersN > 0)
        {
            var folders = savefileto_util.getUnicharPref(this.prefs, this.folderPrefPrefix + "recentFolders");
            folders = savefileto_util.parseFolders(folders, false);
            recentFolders.push(folder);
            for (var i = 0; i < folders.length; i++)
                if (folders[i].name != folder.name)
                    recentFolders.push(folders[i]);
            if (recentFolders.length > this.recentFoldersN)
                recentFolders.splice(this.recentFoldersN, recentFolders.length - this.recentFoldersN);
        }
        recentFolders = savefileto_util.foldersToString(recentFolders);
        savefileto_util.setUnicharPref(this.prefs, this.folderPrefPrefix + "recentFolders", recentFolders);
    },
    
    getDownloadPrefs: function()
    {
        var p = savefileto_util.getPrefBranch("browser.download.");
        var prefs = {};
        prefs.useDownloadDir = p.getBoolPref("useDownloadDir");
        prefs.folderList = p.getIntPref("folderList");
        try
        {
            prefs.dir = p.getComplexValue("dir", Components.interfaces.nsILocalFile);
        }
        catch (e if e.result === Components.results.NS_ERROR_UNEXPECTED) { /* ignore */ }
        try
        {
            prefs.lastDir = p.getComplexValue("lastDir", Components.interfaces.nsILocalFile);
        }
        catch (e if e.result === Components.results.NS_ERROR_UNEXPECTED) { /* ignore */ }
        return prefs;
    },
    
    setDownloadPrefs: function(prefs)
    {
        var p = savefileto_util.getPrefBranch("browser.download.");
        // in ff2, donwload manager window listens to pref changes, so the order here is important
        if (prefs.dir)
        {
            p.setComplexValue("dir", Components.interfaces.nsILocalFile, prefs.dir);
            p.setIntPref("folderList", prefs.folderList);
            p.setBoolPref("useDownloadDir", prefs.useDownloadDir);
        }
        else
        {
            p.setBoolPref("useDownloadDir", prefs.useDownloadDir);
            p.setIntPref("folderList", prefs.folderList);
            p.clearUserPref("dir");
        }
        if (prefs.lastDir)
            p.setComplexValue("lastDir", Components.interfaces.nsILocalFile, prefs.lastDir);
        else
            p.clearUserPref("lastDir");
    }
};


/*
 * http://kb.mozillazine.org/Firefox_:_FAQs_:_About:config_Entries#Browser..2A
 * 
 * browser.download
 *  .useDownloadDir        bool  false: prompt; true: no prompt, download to folder defined in .folderList
 *  .folderList            int   0: desktop; 1: downloads folder; 2: folder in .dir
 *  .dir                   str   user defined download folder
 *  .lastDir               str   last folder used in Save Xxx As...
 *  .save_converter_index  int   the format to save web pages in (Save Page As...)
 *
 * .lastDir does not affect downloading but it is overwritten after it.
 * The original value is preserved because we don't want to affect the normal
 * behavior of Save Xxx As....
 * .save_converter_index on the other hand is always prompted when Save Page to
 * is used (skipPrompt has no effect) so there is no point in keeping the original value.
 *
 * See toolkit.jar:content/global/contentAreaUtils.js:getTargetFile() for actual implementation.
 *
 * Note that ff3's saveLink() doesn't honor the above preferences.
 * See savefileto_saveLink_prompt() below for details.
 */

// based on browser.jar:content/browser/browser.js:nsContextMenu.saveImage()  (context menu > Save Image as...)
function savefileto_saveImage(e, folder)
{
    // ff3.1+ uses mediaURL
    var imageUrl = (gContextMenu.mediaURL ? gContextMenu.mediaURL : gContextMenu.imageURL);
    // ff3 has no docURL
    var doc = gContextMenu.target.ownerDocument;
    var docUri = (gContextMenu.docURL ? makeURI(gContextMenu.docURL) : doc.documentURIObject);
    var principal = (gContextMenu.docURL ? gContextMenu.docURL : doc.nodePrincipal);
    urlSecurityCheck(imageUrl, principal);
    var prefsBackup = this.getDownloadPrefs();
    try
    {
        var prompt = e.ctrlKey || this.prompt;
        var prefs = {useDownloadDir: (! prompt), folderList: 2, dir: folder.file, lastDir: folder.file};
        this.setDownloadPrefs(prefs);
        // toolkit.jar:content/global/contentAreaUtils.js
        // saveImageURL(aURL, aFileName, aFilePickerTitleKey, aShouldBypassCache, aSkipPrompt, aReferrer)
        saveImageURL(imageUrl, null, "SaveImageTitle", false, (! prompt), docUri);
        // Note that execution continues even if the save was cancelled.
        // saveImageURL() doesn't return anything so it's impossible to tell.
        this.addToRecentFolders(folder);
    }
    finally
    {
        // restore original browser download preferences
        this.setDownloadPrefs(prefsBackup);
    }
};

/* 
 * It turns out that there's no "built-in" Content-Disposition filename support
 * in ff2. This is fixed in bug 299372 for ff3, but it only handles situations 
 * where filename is prompted, so no-prompt download must be made by hand. 
 */

function savefileto_saveLink(e, folder)
{
    var prompt = e.ctrlKey || this.prompt;
    if (prompt)
        savefileto_saveLink_prompt.apply(this, arguments);
    else
        savefileto_saveLink_noPrompt.apply(this, arguments);
}

function savefileto_saveLink_prompt(e, folder)
{
    var prefsBackup = this.getDownloadPrefs();
    try
    {
        var prompt = true;
        var prefs = {useDownloadDir: (! prompt), folderList: 2, dir: folder.file, lastDir: folder.file};
        this.setDownloadPrefs(prefs);
        // browser.jar:content/browser/browser.js:nsContextMenu.saveLink()  (context menu > Save Link as...)
        gContextMenu.saveLink();
        // Note that execution continues even if the save was cancelled.
        // saveURL() doesn't return anything so it's impossible to tell.
        this.addToRecentFolders(folder);
    }
    finally
    {
        // ff3 & saveLink():
        //  - doesn't honor |browser.download.*| prefs, it uses always |lastDir|
        //  - download dir is determined after saveLink() has returned, so our change-saveLink()-restore workaround doesn't work
        //
        // So the only option is to not restore |lastDir| to its original value.
        if (savefileto_util.compareVersion(savefileto_util.getBrowserVersion(), "3.0") >= 0)
            prefsBackup.lastDir = prefs.lastDir;
        
        // restore original browser download preferences
        this.setDownloadPrefs(prefsBackup);
    }
}

function savefileto_saveLink_noPrompt(e, folder)
{
    var linkUrl = gContextMenu.linkURL;
    // ff3 has no docURL
    var doc = gContextMenu.target.ownerDocument;
    var docUri = (gContextMenu.docURL ? makeURI(gContextMenu.docURL, doc.characterSet) : doc.documentURIObject);
    var principal = (gContextMenu.docURL ? gContextMenu.docURL : doc.nodePrincipal);
    urlSecurityCheck(linkUrl, principal);
    
    var ioSvc = Components.classes["@mozilla.org/network/io-service;1"].getService(Components.interfaces.nsIIOService);
    
    // source url
    var sourceUri = gContextMenu.getLinkURI();
    sourceUri.QueryInterface(Components.interfaces.nsIURL);
    
    // target dir, file, uri
    var targetDir = folder.file;
    var targetFileName = getDefaultFileName(null, sourceUri, docUri, null);
    var targetFile = savefileto_getTargetFile(targetDir, targetFileName);
    var targetUri = ioSvc.newFileURI(targetFile);
    
    // nsIWebProgressListener to listen possible content-disposition
    function WebProgressListener()
    {
        this.creationTime = Date.now();
    };
    
    WebProgressListener.prototype =
    {
        // nsISupports
        QueryInterface: function(iid)
        {
            if (iid.equals(Components.interfaces.nsIWebProgressListener) ||
                iid.equals(Components.interfaces.nsIWebProgressListener2) ||
                iid.equals(Components.interfaces.nsISupports))
                return this;
            throw Components.results.NS_ERROR_NO_INTERFACE;
        },
        
        // nsIWebProgressListener2
        onProgressChange64: function(webProgress, request, curSelfProgress, maxSelfProgress, curTotalProgress, maxTotalProgress) {},
        onRefreshAttempted: function(webProgress, refreshURI, millis, sameURI) {},
        
        // nsIWebProgressListener
        onLocationChange: function(webProgress, request, location) {},
        onProgressChange: function(webProgress, request, curSelfProgress, maxSelfProgress, curTotalProgress, maxTotalProgress) {},
        onSecurityChange: function(webProgress, request, state) {},
        onStatusChange: function(webProgress, request, status, message) {},
        
        onStateChange: function(webProgress, request, stateFlags, status)
        {
            this._onStateChange(webProgress, request, stateFlags, status);
            if (this.transfer)
                this.transfer.onStateChange(webProgress, request, stateFlags, status);
        },
        
        _onStateChange: function(webProgress, request, stateFlags, status)
        {
            if (stateFlags & Components.interfaces.nsIWebProgressListener.STATE_START)
            {
                if (this.transfer)
                    return;
                
                var trueTargetUri = targetUri;
                
                try
                {
                    // extract content-disposition filename parameter if there is one
                    var channel = request.QueryInterface(Components.interfaces.nsIHttpChannel);
                    var contentDisposition = channel.getResponseHeader("Content-Disposition");
                    var trueFilename = Components.classes["@mozilla.org/network/mime-hdrparam;1"].getService(Components.interfaces.nsIMIMEHeaderParam).getParameter(contentDisposition, "filename", null, false, {});
                    
                    var trueTargetFile = savefileto_getTargetFile(targetDir, trueFilename);
                    trueTargetUri = ioSvc.newFileURI(trueTargetFile);
                    this.trueTargetFilename = trueTargetFile.leafName;
                    this.tempFile = targetFile;
                }
                catch (e if ((e.result == Components.results.NS_ERROR_NO_INTERFACE) || (e.result == Components.results.NS_ERROR_NOT_AVAILABLE) || (e.result == Components.results.NS_ERROR_ILLEGAL_VALUE))) { /* ignore */ }
                
                this.initTransfer(trueTargetUri, this.tempFile);
            }
            else if (stateFlags & Components.interfaces.nsIWebProgressListener.STATE_STOP)
            {
                if (! this.tempFile)
                    return;
                
                // rename temp file to true target file
                this.tempFile.moveTo(null, this.trueTargetFilename);
                delete this.tempFile;
            }
        },
        
        initTransfer: function(trueTargetUri, tempFile)
        {
            if (this.transfer)
                return;
            
            // ff1.5 seems to leak if notify isn't removed
            timer.cancel();
            delete callback.notify;
            
            this.transfer = Components.classes["@mozilla.org/transfer;1"].createInstance(Components.interfaces.nsITransfer);
            var displayName = null;
            var mimeInfo = null;
            var time = this.creationTime * 1000;
            this.transfer.init(sourceUri, trueTargetUri, displayName, mimeInfo, time, tempFile, persist);
            
            // nsIWebProgressListener2.onRefreshAttempted is ff3 only
            for each (var name in ["onLocationChange", "onProgressChange", "onSecurityChange", "onStatusChange", "onProgressChange64", "onRefreshAttempted"])
                if (this.transfer[name])
                    this[name] = savefileto_util.closure(this.transfer, this.transfer[name]);
        }
    }
    
    // nsIWebBrowserPersist
    var persist = Components.classes["@mozilla.org/embedding/browser/nsWebBrowserPersist;1"].createInstance(Components.interfaces.nsIWebBrowserPersist);
    var progressListener = new WebProgressListener(persist);
    persist.progressListener = progressListener;
    persist.persistFlags =
        Components.interfaces.nsIWebBrowserPersist.PERSIST_FLAGS_AUTODETECT_APPLY_CONVERSION |
        Components.interfaces.nsIWebBrowserPersist.PERSIST_FLAGS_BYPASS_CACHE |
        Components.interfaces.nsIWebBrowserPersist.PERSIST_FLAGS_REPLACE_EXISTING_FILES;
    
    // timeout for waiting the HTTP headers
    // otherwise it would be impossible to cancel the download from the ui
    var timer = Components.classes["@mozilla.org/timer;1"].createInstance(Components.interfaces.nsITimer);
    var callback = {notify: function() { progressListener.initTransfer(targetUri, null); }};
    var timeout = this.prefs.getIntPref("responseHeaderTimeout");
    timer.initWithCallback(callback, timeout, Components.interfaces.nsITimer.TYPE_ONE_SHOT);
    
    var referer = doc.documentURIObject;
    persist.saveURI(sourceUri, null, referer, null, null, targetUri);
    
    this.addToRecentFolders(folder);
}

function savefileto_savePage(e, folder)
{
    var prefsBackup = this.getDownloadPrefs();
    try
    {
        var prefs = {useDownloadDir: (! prompt), folderList: 2, dir: folder.file, lastDir: folder.file};
        this.setDownloadPrefs(prefs);
        // toolkit.jar:content/global/contentAreaUtils.js
        // saveDocument(aDocument, aSkipPrompt)
        saveDocument(window.content.document);
        // Note that execution continues even if the save was cancelled.
        // saveDocument() doesn't return anything so it's impossible to tell.
        this.addToRecentFolders(folder);
    }
    finally
    {
        // restore original browser download preferences
        this.setDownloadPrefs(prefsBackup);
    }
};

function savefileto_getTargetFile(dir, filename)
{
    var file = dir.clone();
    file.append(filename);
    
    // copy paste from toolkit.jar:content/global/contentAreaUtils.js:getTargetFile()
    var collisionCount = 0;
    while (file.exists()) {
        collisionCount++;
        if (collisionCount == 1) {
            // Append "(2)" before the last dot in (or at the end of) the filename
            // special case .ext.gz etc files so we don't wind up with .tar(2).gz
            if (file.leafName.match(/\.[^\.]{1,3}\.(gz|bz2|Z)$/i))
                file.leafName = file.leafName.replace(/\.[^\.]{1,3}\.(gz|bz2|Z)$/i, "(2)$&");
            else
                file.leafName = file.leafName.replace(/(\.[^\.]*)?$/, "(2)$&");
        }
        else {
            // replace the last (n) in the filename with (n+1)
            file.leafName = file.leafName.replace(/^(.*\()\d+\)/, "$1" + (collisionCount+1) + ")");
        }
    }
    
    return file;
}


window.addEventListener("load", function(e) { savefileto.doOnLoad(e); }, false);
